@@ -12,7 +12,7 @@
 #include <deque>
 #include <array>
 #include <tuple>
-#if __cplusplus >= 201703L
+#if TOML11_CPLUSPLUS_STANDARD_VERSION >= 201703L
 #include <string_view>
 #endif
 
@@ -224,7 +224,7 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_get_string_type, value_type, test_value_types
         BOOST_TEST("foobar" == x);
     }
 
-#if __cplusplus >= 201703L
+#if TOML11_CPLUSPLUS_STANDARD_VERSION >= 201703L
     {
         value_type v("foo", toml::string_t::basic);
         BOOST_TEST("foo" == toml::get<std::string_view>(v));
